module mpi_task_list ! simple fifo list
  use my_data_types
  implicit none

  private

  public list_element
  type list_element
    type(task_data) :: dat
    integer :: tag
    type(list_element), pointer :: next
  end type list_element

  public list
  type list
    type(list_element), pointer :: first
  end type list

  public :: list_create
  public :: list_delete
  public :: list_push
  public :: list_dequeue
  public :: list_is_empty
  public :: list_num_elements
contains

  subroutine list_create(l)
    implicit none
    type(list), intent(out) :: l

    nullify(l%first)
  end subroutine list_create

  subroutine list_delete(l)
    implicit none
    type(list), intent(inout) :: l
    type(list_element), pointer :: trash

    do while(associated(l%first))
      trash => l%first
      l%first => l%first%next
      deallocate(trash); nullify(trash)
    enddo
  end subroutine list_delete

  subroutine list_push(l,dat,tag)
    implicit none
    type(list), intent(inout) :: l
    type(task_data), intent(in) :: dat
    integer, intent(in) :: tag
    type(list_element), pointer :: runner

    if (list_is_empty(l)) then
      allocate(l%first)
      nullify(l%first%next)
      l%first%dat = dat
      l%first%tag = tag
    else
      runner => l%first
      do while (associated(runner%next))
        runner => runner%next
      enddo
      allocate(runner%next)
      runner => runner%next
      nullify(runner%next)
      runner%dat = dat
      runner%tag = tag
    endif
  end subroutine list_push

  subroutine list_dequeue(l,dat,tag)
    implicit none
    type(list), intent(inout) :: l
    type(task_data), intent(out) :: dat
    integer, intent(out) :: tag
    type(list_element), pointer :: trash

    if (list_is_empty(l)) then
      ! empty list! throw some kind of error! TODO
    else
      dat = l%first%dat
      tag = l%first%tag
      trash => l%first
      l%first => l%first%next
      deallocate(trash); nullify(trash)
    endif
  end subroutine list_dequeue

  function list_is_empty(l)
    implicit none
    type(list), intent(in) :: l
    logical :: list_is_empty
    list_is_empty = .not. associated(l%first)
  end function list_is_empty

  function list_num_elements(l)
    implicit none
    type(list) :: l
    integer :: list_num_elements
    type(list_element), pointer :: runner

    list_num_elements = 0
    if (list_is_empty(l)) return
    runner => l%first
    do while (associated(runner))
      list_num_elements = list_num_elements+1
      runner => runner%next
    enddo
  end function list_num_elements

end module mpi_task_list
